package com.maroontress.coverture;

import com.maroontress.coverture.gcda.DataRecord;
import com.maroontress.coverture.gcda.FunctionDataRecord;
import com.maroontress.coverture.gcno.FunctionGraphRecord;
import com.maroontress.coverture.gcno.NoteRecord;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.channels.FileChannel;
import java.util.Arrays;
import java.util.Collection;
import java.util.TreeMap;

/**
   gcnoեѡ̤ݻޤ
*/
public final class Note {

    /** gcnoեΥСֹǤ */
    private int version;

    /**
       gcnoեΥॹפǤgcdaեƱȤƤ
       뤳Ȥǧ뤿˻Ѥޤ
    */
    private int stamp;

    /**
       gcnoեΥե륷ƥǤΥॹפǤե
       ΥॹפӤ뤿˻Ѥޤ
    */
    private long lastModified;

    /** ؿդȤμ̻ҤΥޥåפǤ */
    private TreeMap<Integer, FunctionGraph> map;

    /**
       Ρȥ쥳ɤ饤󥹥󥹤ޤ

       @param rec Ρȥ쥳
       @param lastModified gcnoեΥե륷ƥΥॹ
       
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
    */
    private Note(final NoteRecord rec, final long lastModified)
	throws CorruptedFileException {
	version = rec.getVersion();
	stamp = rec.getStamp();
	this.lastModified = lastModified;
	map = new TreeMap<Integer, FunctionGraph>();

	FunctionGraphRecord[] list = rec.getList();
	for (FunctionGraphRecord e : list) {
	    FunctionGraph fg = new FunctionGraph(e);
	    map.put(fg.getId(), fg);
	}
    }

    /**
       ΡȤXMLǽϤޤ

       @param out 
    */
    public void printXML(final PrintWriter out) {
	out.printf("<note version='0x%x' stamp='0x%x' lastModified='%d'>\n",
		   version, stamp, lastModified);
	Collection<FunctionGraph> allGraphs = map.values();
	for (FunctionGraph g : allGraphs) {
	    g.printXML(out);
	}
	out.printf("</note>\n");
    }

    /**
       ǡ쥳ɤꤷޤ

       @param rec ǡ쥳
       @param lastModified gcdaեΥե륷ƥΥॹ
       
       @throws CorruptedFileException
    */
    private void setDataRecord(final DataRecord rec, final long lastModified)
	throws CorruptedFileException {
	if (version != rec.getVersion()) {
	    throw new CorruptedFileException("gcda file: version mismatch.");
	}
	if (stamp != rec.getStamp()) {
	    throw new CorruptedFileException("gcda file: timestamp mismatch.");
	}
	if (this.lastModified > lastModified) {
	    System.out.println("warning: gcno file is newer than gcda file.");
	}
	FunctionDataRecord[] list = rec.getList();
	for (FunctionDataRecord e : list) {
	    int id = e.getId();
	    FunctionGraph g = map.get(id);
	    if (g == null) {
		System.out.printf("warning: unknown function id '%d'.", id);
		continue;
	    }
	    g.setFunctionDataRecord(e);
	}
    }

    /**
       gcdaեѡơΡȤ˥󥿤ɲäޤ
       ͥޥåפΤǡ2GХȤĶեϰޤ

       եƤʾϡɸ२顼Ϥ˥åȥ졼
       Ϥޤ

       @param path gcdaեΥѥ
    */
    private void parseData(final String path) {
	if (!path.endsWith(".gcda")) {
	    System.err.printf("%s: suffix is not '.gcda'.\n", path);
	    return;
	}
	File file = new File(path);
	try {
	    FileChannel ch = new RandomAccessFile(file, "r").getChannel();
	    ByteBuffer bb = ch.map(FileChannel.MapMode.READ_ONLY,
				   0, ch.size());
	    try {
		DataRecord dataRecord = new DataRecord(bb);
		setDataRecord(dataRecord, file.lastModified());
	    } catch (UnexpectedTagException e) {
		e.printStackTrace();
	    } catch (CorruptedFileException e) {
		e.printStackTrace();
	    } finally {
		ch.close();
	    }
	} catch (FileNotFoundException e) {
	    System.err.println(path + ": not found");
	} catch (IOException e) {
	    e.printStackTrace();
	}
    }

    /**
       gcnoեѡơΡȤޤͥޥå
       Τǡ2GХȤĶեϰޤ

       եƤʾϡɸ२顼Ϥ˥åȥ졼
       Ϥơnull֤ޤ

       @param path gcnoեΥѥ
       @return Ρ
       @throws IOException
    */
    public static Note parse(final String path) throws IOException {
	if (!path.endsWith(".gcno")) {
	    System.err.printf("%s: suffix is not '.gcno'.", path);
	    return null;
	}
	File file = new File(path);
	FileChannel ch = new RandomAccessFile(file, "r").getChannel();
	ByteBuffer bb = ch.map(FileChannel.MapMode.READ_ONLY, 0, ch.size());
	Note note = null;

	try {
	    NoteRecord noteRecord = new NoteRecord(bb);
	    note = new Note(noteRecord, file.lastModified());
	} catch (UnexpectedTagException e) {
	    e.printStackTrace();
	    return null;
	} catch (CorruptedFileException e) {
	    e.printStackTrace();
	    return null;
	} finally {
	    ch.close();
	}
	note.parseData(path.substring(0, path.lastIndexOf('.')) + ".gcda");
	return note;
    }
}
