package com.maroontress.coverture.gcda;

import com.maroontress.coverture.CorruptedFileException;
import com.maroontress.coverture.Tag;
import com.maroontress.coverture.UnexpectedTagException;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.ArrayList;
import java.util.Arrays;

/**
   ǡ쥳ɤǤ

   data: int32:magic int32:version int32:stamp
         {function-data* object-summary program-summary*}*
*/
public final class DataRecord {

    /** ӥåǥΥޥåʥСǤ */
    private static final byte[] MAGIC_BE = {'g', 'c', 'd', 'a'};

    /** ȥ륨ǥΥޥåʥСǤ */
    private static final byte[] MAGIC_LE = {'a', 'd', 'c', 'g'};

    /**
       gcdaեΥСֹǤ
    */
    private int version;

    /**
       gcdaեΥॹפǤgcnoեƱȤƤ
       뤳Ȥǧ뤿˻Ѥޤ
    */
    private int stamp;

    /** ؿǡ쥳ɤΥꥹȤǤ */
    private ArrayList<FunctionDataRecord> list;

    /** ֥ȤΥޥ쥳ɤǤ */
    private SummaryRecord objectSummary;

    /** ץΥޥ쥳ɤΥꥹȤǤ */
    private ArrayList<SummaryRecord> programSummaries;

    /**
       ХȥХåեFUNCTION_DATA쥳ɤϤб륤
       󥹤ơꥹȤɲäޤХȥХåեΰ֤
       FUNCTION_DATA쥳ɤƬΰ֤ǤʤФʤޤ󡣥ХȥХ
       եΰ֤FUNCTION_DATA쥳ɤμ˿ʤߤޤ

       OBJECT_SUMMARYϤ-1֤ޤOBJECT_SUMMARY
       ϤϥХȥХåեΰ֤OBJECT_SUMMARYľ
       ˰ưޤ

       @param bb ХȥХåե
       @return FUNCTION_DATA쥳ɤϤ0Ǥʤ-1
       @throws IOException ϥ顼
       @throws CorruptedFileException եι¤Ƥ뤳Ȥ򸡽
       @throws UnexpectedTagException ͽʤ򸡽
    */
    private int parseFunctionData(final ByteBuffer bb)
	throws IOException, UnexpectedTagException, CorruptedFileException {
	int saved = bb.position();
	int tag = bb.getInt();
	switch (tag) {
	default:
	    String m = String.format("unexpected tag: 0x%x", tag);
	    throw new UnexpectedTagException(m);
	case Tag.OBJECT_SUMMARY:
	    return -1;
	case Tag.FUNCTION:
	    list.add(new FunctionDataRecord(bb));
	    break;
	}
	return 0;
    }

    /**
       ХȥХåեgcdaեѡơǡ쥳ɤ
       ޤХȥХåեΰ֤ϥХåեƬǤʤФʤ
       ϥХȥХåեΰ֤Ͻü˰ưޤ

       @param bb ХȥХåե
       @throws IOException
       @throws CorruptedFileException
       @throws UnexpectedTagException
    */
    public DataRecord(final ByteBuffer bb)
	throws IOException, UnexpectedTagException, CorruptedFileException {
	byte[] magic = new byte[4];
	bb.get(magic);
	if (Arrays.equals(magic, MAGIC_BE)) {
	    bb.order(ByteOrder.BIG_ENDIAN);
	} else if (Arrays.equals(magic, MAGIC_LE)) {
	    bb.order(ByteOrder.LITTLE_ENDIAN);
	} else {
	    throw new CorruptedFileException();
	}
	version = bb.getInt();
	stamp = bb.getInt();
	list = new ArrayList<FunctionDataRecord>();
	programSummaries = new ArrayList<SummaryRecord>();

	while (parseFunctionData(bb) == 0) {
	    continue;
	}
	objectSummary = new SummaryRecord(bb);
	while (bb.hasRemaining()) {
	    int saved = bb.position();
	    int tag = bb.getInt();
	    switch (tag) {
	    default:
		String m = String.format("unexpected tag: 0x%x", tag);
		throw new UnexpectedTagException(m);
	    case Tag.FUNCTION:
		bb.position(saved);
		return;
	    case Tag.PROGRAM_SUMMARY:
		programSummaries.add(new SummaryRecord(bb));
		break;
	    }
	}
    }

    /**
       Сޤ

       @return С
    */
    public int getVersion() {
	return version;
    }

    /**
       ॹפޤ

       @return ॹ
    */
    public int getStamp() {
	return stamp;
    }

    /**
       ؿǡ쥳ɤޤ

       @return ؿǡ쥳ɤ
    */
    public FunctionDataRecord[] getList() {
	return list.toArray(new FunctionDataRecord[0]);
    }

    /**
       ֥ȤΥޥ쥳ɤޤ

       @return ֥ȤΥޥ쥳
    */
    public SummaryRecord getObjectSummary() {
	return objectSummary;
    }

    /**
       ץΥޥ쥳ɤޤ

       @return ץΥޥ쥳ɤ
    */
    public SummaryRecord[] getProgramSummaries() {
	return programSummaries.toArray(new SummaryRecord[0]);
    }
}
